' Attiny-45-Morseuhr Version 1.0 slim
'
' Gibt alle 15 min. die Minutenzahl aus und zur vollen Stunde die Stundenzahl
' ("es ist X Uhr"),
' ausserdem kann die Zeit durch Eingabe von "T" unmittelbar abgefragt werden.
'
' Befehlsliste:
'
' ?     Auflistung aller Befehle
' Z     Zeit setzen
' T     Zeit abfragen
' G 1/0   Gong (Schlagwerk) ein/aus
' M     Morsegeschwindigkeit setzen
' C     Check -> Ausgabe des Gongstatus
'
' Pinbelegung:
'
' PB0 Buzzer
' PB1 Punktkontakt
' PB2 Strichkontakt
' PB3 und PB4: Quarz


' Quarz schwingt mit 3,686400 MZz; Oszillator wird duch 8 geteilt (Takt ist dann 460800 Hz).
'
' Timer1- Vorteiler teilt durch 2048, Timer 1 dann durch 225 -> Sekundenimpulse
'
' Nach 60 Interrupts ist eine Minute um; mit Minuten wird weitergerechnet.
'
' Da der Quarz ohne Brdekondensatoren betrieben wird, ist die Uhr grundstzlich etwas zu schnell.
' Ausgleich durch Wait- Befehl in der Interrupt- Routine ( 110 sec * 86400 sec/Tag  = 9,5 sec/Tag ).

' Fusebits: lfuse: 0x7d   hfuse: 0xdf  (Taktquelle Quarzoszillator, Teiler 1/8 , kein Brown-Out)
'
'
' Bugfix Version 1.01: Geschwindigkeits- Feineinstellung in der Interrupt- Routine funktionierte nicht.
' Die Wartezeit hatte keine Auswirkung, weil mit  TCNT1 = 31 zwar den 8-Bit-Teiler, presetted wird,
' whrend der 200 s Wartezeit lief aber der Vorteiler einrfach weiter. Reset Vorteiler: GTCCR.PSR1 = 1
'
'
' Bugfix Version 1.02: GTCCR.PSR1 = 1 fhrt dazu, da die Uhr stets zu langsam luft (Bascom verbraucht
' Zeit fr den Sprung in die Interrupt-Routine). Ausgleich erfolgt durch Vorstellen der Uhr um einige
' Sekunden beim Tageswechsel
'
'----------------------------------------------------------


$regfile = "Attiny45.dat"
$crystal = 460800                                           ' 3,6864 MHz / 8

$hwstack = 32
$swstack = 8
$framesize = 24


' Variablen fr Zeitberechnung

Dim S As Word                                               ' Sekunden
Dim T As Word                                               ' Tagesminuten 0 ..... 1440
Dim Mi As Word                                              ' Minuten- Rest zur vollen Stunde
Dim V As Word                                               ' Minuten- Rest zur vollen Viertelstunde
Dim H As Word                                               ' Stunden
Dim Zh As Word                                              ' Zehnerstelle Stunden
Dim Eh As Word                                              ' Einerstelle Stunden
Dim Zmi As Word                                             ' Zehnerstelle Minuten
Dim Emi As Word                                             ' Einerstelle Minuten
Dim Hilf As Word                                            ' Hilfsgroesse fr Tagesminutenberechnung
Dim Gong As Byte                                            ' Ein- / Ausschalter Gong


' Variablen fr Morsezeichen- Erzeugung

Dim I As Byte                                               ' wird in Zhlschleife verwendet
Dim N As Word                                               ' Aktueller Morsebuchstabe
Dim M As Byte                                               ' wird frs Auslesen der Morsebits verwendet
Dim L As Byte                                               ' wird fr niederwertigstes Bit von M verwendet


' Variablen fr Morsezeichen- Decodierung

Dim E As Byte                                               ' resultierendes ACCII-Byte der Decodierroutine
Dim D As Byte                                               ' Morsebyte in der Decodierroutine
Dim K As Byte                                               ' Byte, in dem eine "1" geshiftet wird
Dim J As Byte                                               ' Zhlvariable fr tolerante Pausenprfung
Dim P As Byte                                               ' Punktspeicher-Bit der Decodierroutine


' Variablen fr Ziffern- Eingabe

Dim G(4) As Byte                                            ' nimmt die Ziffern auf
Dim R As Byte                                               ' Anzahl angeforderter Ziffern
Dim W As Byte                                               ' Zhler fr Eingabewiederholung der Ziffern
Dim Notbremse As Byte                                       ' Timer fr Ausstieg aus unvollstndigen Eingaben

Dim Txtstring As String * 52
Dim Txt(52) As Byte At Txtstring Overlay


' Variablen fr Morsezeichen- Timing

Dim Punktlaenge1 As Word                                    ' Punktlnge in "Empfangsrichtung"
Dim Punktlaenge2 As Word                                    ' Punktlnge in "Senderichtung"
Dim Punktachtel As Word
Dim Strichlaenge1 As Word                                   ' Strichlnge
Dim Strichlaenge2 As Word
Dim Wortpause1 As Word                                      ' Lnge Wortpause
Dim Wortpause2 As Word


Mcucr.sm0 = 0                                               ' mcucr  fr Idle konfigurieren
Mcucr.sm1 = 0                                               ' mcucr  Idle zweites Bit

Pcmsk.1 = 1                                                 ' PCINT fr PB1 freigeben
Pcmsk.2 = 1                                                 ' PCINT fr PB2 freigeben
Gimsk.pcie = 1                                              ' PCints aktivieren
Sreg.7 = 1                                                  ' Interrupts generell freigeben

Portb = &B0000110                                           ' Pullups fr Strich- und Punktkontakt
Ddrb = &B00000000                                           ' erstmal alles Eingang


' Stromspar- Funktionen nutzen

Didr0 = &B00011001                                          ' digitale Eingnge ausser PB1 und PB2 abschalten
Acsr.acd = 0                                                ' Analog-Komparator ausschalten
Prr.prusi = 1                                               ' USI abschalten (power reduction  usi)
Prr.pradc = 1                                               ' ADC abschalten (power reduction  adc)



' Morseton mit Timer0 erzeugen:
'
' Timer0 zhlt das Register TCNT0 hoch. Es wird mit Register OCR0A verglichen.
' Wenn TCNT0 = OCR0A, wird Output OC0A/PB0 umgeschaltet (getoggled)
' Timer0 erzeugt so Dauerton; er wird jedoch nur an den Ausgangspin durchgeschaltet,
' wenn das Datenrichtungs- Register DDRB.0 = 1 ist. Bei 0 wird der Ton unterdrckt.
'
Config Timer0 = Timer , Prescale = 64 , Compare A = Toggle , Clear Timer = 1


' 1- Sekunden- Interrupts mit Timer 1 erzeugen
'
Config Timer1 = Timer , Prescale = 2048
Enable Timer1

On Timer1 Ontimer1                                          ' Interrupt-Routine fr Timer1-Overflow


Const Tonhoehe1 = 12
Const Tonhoehe2 = 18


Buzzer Alias Ddrb.0
Punktkontakt Alias Pinb.1
Strichkontakt Alias Pinb.2


'----------------------------------------------------------



Init:


'---  vorlufige Werte -------

  Punktlaenge1 = 80                                         ' Punktlnge in "Empfangsrichtung"
  Punktlaenge2 = 100                                        ' Punktlnge in "Senderichtung"
  Punktachtel = 10
  Strichlaenge1 = Punktlaenge1 * 3                          ' Strichlnge
  Strichlaenge2 = Punktlaenge2 * 3
  Wortpause1 = Punktlaenge1 * 5                             ' Lnge Wortpause
  Wortpause2 = Punktlaenge2 * 5

  Gong = 1

  Gosub Zeitausgabe                                         ' Kontrollausgabe nach Reset




'--- Hauptschleife -----------

Do                                                          ' Hauptschleife


       Prr.prtim0 = 0                                       ' Timer0 anschalten
       Mcucr.se = 0                                         ' Sleeepmode verbieten
       Mcucr.pcie = 0                                       ' Tasteninterrupts verbieten


       If T > 1439 Then                                     ' Prfen auf Tageswechsel
'         Wait 5                                              ' Ausgleich bei zu schneller Uhr
'         S = 0
         S = 4                                              ' Ausgleich bei zu langsamer Uhr
         T = 0
       End If

       Gosub Zeitberechnung

       If Gong = 1 Then
          Gosub Gong
       End If

       Gosub Menue
'                                                             Idle- Modus vorbereiten
       Prr.prtim0 = 1                                       ' Timer0 abschalten
       Mcucr.pcie = 1                                       ' Pin Change Interrupt enable
       Mcucr.se = 1                                         ' Sleep enable
       !SLEEP

Loop




'------- Zeitberechnungen und Ausgabe der Zeiten ---------


Zeitberechnung:

       H = T / 60                                           ' Uhrzeit auf Stunden gerundet
       Mi = T Mod 60                                        ' Minuten- Rest  (zur vollen Stunde)
       Zh = H / 10                                          ' Zehnerstelle Stunden
       Eh = H Mod 10                                        ' Einerstelle Stunden
       Zmi = Mi / 10                                        ' Einerstelle Minuten
       Emi = Mi Mod 10                                      'Einerstelle Minuten
       V = T Mod 15                                         ' V = 0 bei vollen Viertelstunden
Return



Gong:

      If S < 2 Then                                         ' unterdrckt mehrfache Ausgabe
         If V = 0 Then                                      ' volle Viertelstunde
            If Mi = 0 Then                                  ' Sonderfall: volle Stunde
               Txtstring = " ES IST "
               Gosub Sendstring
               If Zh > 0 Then                               ' wenn zweistellige Stundenzahl
                  M = Zh                                    ' Zehnerstunde senden
                  Gosub Sendnumber
               End If
               M = Eh                                       ' Einerstunde senden
               Gosub Sendnumber
               Txtstring = " UHR "
               Gosub Sendstring
            Else                                            ' zur 15., 30., 45. Minute nur Minutenausgabe
               M = Zmi                                      ' Zehnerminute
               Gosub Sendnumber
               M = Emi                                      ' Einerminute
               Gosub Sendnumber
            End If
         End If
      End If
Return



Zeitausgabe:


        Txtstring = " ES IST "
        Gosub Sendstring

        If Zh > 0 Then
           M = Zh                                           ' Zehnerstelle Stunden
           Gosub Sendnumber
        End If
        M = Eh                                              ' Einerstelle Stunden
        Gosub Sendnumber

        Txtstring = " UHR "
        Gosub Sendstring

        If Zmi > 0 Then
           M = Zmi                                          ' Zehnerstelle Minuten
           Gosub Sendnumber
        End If
        M = Emi                                             ' Einerstelle Minuten
        Gosub Sendnumber
        Txtstring = " MIN"
        Gosub Sendstring

Return




'------- Aufbereitung der zu sendenden Morsezeichen ------



Sendnumber:

     M = M + 48                                             ' Ascii- Wert 0: 048 ;  Ascii- Wert 9: 057
     M = Lookup(m , Morsetabelle)                           ' Morsezeichen aus Tabelle holen
     Gosub Morse

Return



Sendstring:

     For N = 1 To Len(txtstring)                            ' Textstring ausgeben
        M = Txt(n)
        M = Lookup(m , Morsetabelle)                        ' Morsezeichen aus Tabelle holen
        Gosub Morse
     Next T

Return




'------- Erzeugung der zu sendenden Morsezeichen -------



 Morse:

      Ocr0a = Tonhoehe1

      If M = 0 Then                                         ' Sonderfall: Leerzeichen;
         Waitms Wortpause1
      Goto Zeichenende
      End If

      For I = 1 To 8
         If M = 1 Then                                      ' Das Byte hat nur noch den Wert 1; Zeichenende!
            Goto Zeichenende
         End If
         L = M And &B00000001                               ' niederwertigstes Bit lesen
         Buzzer = 1
         If L = 1 Then
            Waitms Strichlaenge1                            ' ist das Bit 1 -> dah
         Else
            Waitms Punktlaenge1                             ' ist das Bit 0 -> dit
         End If
         Buzzer = 0
         Waitms Punktlaenge1                                ' Pause innerhalb des Morsezeichens
         Shift M , Right , 1                                ' Bits um eine Stelle nach rechts shiften
      Next I

  Zeichenende:
      Waitms Strichlaenge1                                  ' Pause zwischen Morsezeichen

 Return





'--- Hier beginnt die Auswertung der Morse- Eingaben --------



'------ Men ( Befehle bestehen aus 1 Zeichen ) -----


Menue:


  Gosub Decodekeyer

  Select Case E

       Case 71                                              ' 71 : Ascii - Wert "G"
       Txtstring = " GONG?"
       Gosub Sendstring
       R = 1
       Gosub Readnumbers
       If G(1) = 1 Then
           Gong = 1
         Else
           Gong = 0
       End If

     Case 67                                                ' 67 : Ascii- Wert "C"
       Txtstring = " GONG "
       Gosub Sendstring
       M = Gong
       Gosub Sendnumber

      Case 77                                               ' 77 : Ascii - Wert "M"
       Txtstring = " WPM?"
       Gosub Sendstring
       Gosub Speed

     Case 84                                                ' 84 : Ascii - Wert "T"
       Gosub Zeitausgabe

     Case 90                                                ' 90 : Ascii - Wert "Z"
       Txtstring = " ZEIT?"
       Gosub Sendstring
       Gosub Zeitsetzen

     Case 63                                                ' 63 : Ascii - Wert "?"
       Txtstring = " BEFEHLE: CHECK  GONG  MORSESPEED  TIME  ZEITSETZEN"
       Gosub Sendstring

  End Select

Return




' ----- hier folgen die Subroutinen, die durch das Men aufgerufen werden ------


Zeitsetzen:

    S = 0                                                   ' Sekunden auf 0 setzen

    R = 4
    Gosub Readnumbers

    Hilf = G(3) * 10                                        ' G(3) Zehnerstelle Minuten
    T = G(4) + Hilf                                         ' G(4) Einerstelle Minuten
    Hilf = G(2) * 60                                        ' G(2) Einerstelle Stunden
    T = T + Hilf
    Hilf = 600 * G(1)                                       ' G(1) Zehnerstelle Stunden
    T = T + Hilf                                            ' T Tagesminuten

    If T > 1439 Then                                        ' spter als 23.59 Uhr
       T = 1                                                ' Workaround, damit der Gong nicht anspricht
       Txtstring = "RPT"
       Gosub Sendstring
    End If

Return




Speed:

    R = 2
    Gosub Readnumbers

    ' Aus dem abgefragten Wert neue Geschwindigkeit berechnen

Calcspeed:

      G(1) = 10 * G(1)                                      ' Zehnerstelle WPM
      G(1) = G(1) + G(2)                                    ' Einerstella addieren
      Punktlaenge1 = 1320 / G(1)                            ' aus WPM Punktlnge errechnen
      Strichlaenge1 = Punktlaenge1 * 3
      Wortpause1 = Punktlaenge1 * 5
      Punktlaenge2 = 3 * Punktlaenge1                       ' Sendegeschwindigkeit soll nur 2/3 der
      Punktlaenge2 = Punktlaenge2 / 2                       ' Empfangsgeschwindigkeit sein
      Punktachtel = Punktlaenge2 / 8
      If Punktachtel = 0 Then                               ' Mindestwert 1 fr tolerante Pausenprfung
         Punktachtel = 1
      End If
      Strichlaenge2 = Punktlaenge2 * 3
      Wortpause2 = Punktlaenge2 * 5

      If G(1) < 10 Or G(1) > 30 Then                        ' auf plausiblen Wert prfen
          G(1) = 2
          G(2) = 0
          Goto Calcspeed                                    ' Neuberechnung mit den Default- Werten
      End If

Return




'------ Morseingabe- Auswerteroutinen -------


Readnumbers:

  I = 0
  Notbremse = 0

  Do
    If Punktkontakt = 0 Or Strichkontakt = 0 Then           ' Punkt - oder Strichkontakt gedrckt
       I = I + 1
       Gosub Decodekeyer
       G(i) = E                                             ' jeweils den Ascii- Wert des Morsezeichens zuweisen
       Select Case G(i)
         Case 48 To 57                                      ' auf Zahleneingabe prfen
           G(i) = G(i) - 48                                 ' von Ascii in einstellige Zahl umrechnen
         Case Else
           Txtstring = "RPT"                                ' Eingabefehler abfangen
           Gosub Sendstring
           Return
     End Select
    End If

    Notbremse = Notbremse + 1
    Waitms 20
    If Notbremse > 250 Then                                 ' Ausstieg nach lngerer Eingabepause (5 sec)
       Txtstring = "RPT"
       Gosub Sendstring
       Return
    End If
  Loop Until I = R

  Waitms 500

  For W = 1 To R                                            ' zur Kontrolle die Zahlen ausgeben
      M = G(w)
      Gosub Sendnumber
   Next

Return






 '------- Interpretation der eingegebenen Einzel- Morsezeichen  ----

Decodekeyer:                                                ' Prinzip: eine 1 wird in der Variablen "K" von links nach rechts geschoben.
                                                            ' nur wenn Strich gedrckt wurde, wird die 1 in die Variable D bernommen
  Ocr0a = Tonhoehe2
  D = 0
  K = 1
  Do

      If Punktkontakt = 0 Then                              ' Punktkontakt gedrckt
         Buzzer = 1                                         ' Ton an
         Waitms Punktlaenge2
         Buzzer = 0                                         ' Ton aus
         Waitms Punktlaenge2
         Shift K , Left                                     ' Byteinhalt wandert 1 Stelle nach links (von rechts wird eine "0" eingeschoben)
      End If

  Mark:

      If Strichkontakt = 0 Then                             ' Strichkontakt gedrckt
         Buzzer = 1                                         '  Ton an
         Waitms Strichlaenge2
         D = D + K                                          ' die 1 fr den Strich wird in die Variable "D" bernommen
         Shift K , Left                                     ' Byteinhalt wandert 1 Stelle nach rechts, von links eine "0" eingeschoben
         If Punktkontakt = 0 Then                           ' wenn Punktkontakt ebenfalls gedrckt:
            P = 1                                           ' Punktspeicher setzen
         End If
         Buzzer = 0                                         ' Ton aus
         Waitms Punktlaenge2
         If P = 1 Then                                      ' Punktspeicher prfen
            Buzzer = 1                                      ' Ton an
            Waitms Punktlaenge2
            Buzzer = 0                                      ' Ton aus
            Waitms Punktlaenge2
            Shift K , Left                                  ' Byteinhalt wandert 1 Stelle nach links, rechts wird 0 eingeschoben
            P = 0                                           ' Punktspeicher zurcksetzen
           Goto Mark                                        ' Prfen, ob Strichkontakt bereits wieder gedrckt
         End If
      End If

      J = 0
      Do                                                    ' wir warten max. halbe Punktlnge, um unexakte Gebeweise abzufangen
         If Punktkontakt = 0 Or Strichkontakt = 0 Then      ' vorzeitiger Ausstieg, wenn Punktkontakt oder Strichkontakt gedrckt
             Exit Do
         End If
         Waitms Punktachtel
         J = J + 1
      Loop Until J = 4                                      ' Abbruch bei halber Punktlnge

  Loop Until Punktkontakt = 1 And Strichkontakt = 1         ' kein Kontakt mehr gedrckt; Morsezeichen fertig

  D = D + K                                                 ' eine 1 wird als Endbit angehngt
  E = Lookdown(d , Morsetabelle , 91)                       ' inverse Lookup in der Tabelle, um den zugehrigen Ascii- Wert zu ermitteln
  E = E - 1                                                 ' Data- Elemente werden ab 1 gezhlt, Ascii- Tabelle beginnt bei Null

Return




'--- Interrupt- Routine Sekundentakt ------


Ontimer1:                                                   'Timer1-Overflow-Interrupt-Routine

   Gtccr.psr1 = 1                                           ' resettet Vorteiler
   Tcnt1 = 31                                               ' Preset auf 31, damit er 225 Takte bis Ueberlauf zhlt
   S = S + 1
      If S > 59 Then                                        ' Minutenwechsel
        S = 0
        T = T + 1
      End If

Return




'--- Morsezeichen- Tabelle --------------------------


Morsetabelle:

'Ascii 0 - 32

Data 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , &B00101101 , 0 , , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0 , 0
Data 0 , 0 , 0 , 0 , 0 , 0 , 0


' Ascii 32 - 47   ->   _!"#$%&'()*+,-./
Data 0 , 0 , &B01010010 , 0 , 0 , 0 , 0 , 0 , &B00101101 , &B01101101 , 0 , &B00101010 , &B01110011 , &B01100001
Data &B01101010 , &B00101001

'Ascii 48 - 57  ->  0123456789
Data &B00111111 , &B00111110 , &B00111100 , &B00111000 , &B00110000 , &B00100000 , &B00100001 , &B00100011 , &B00100111 , &B00101111

'Ascii 58 - 64  ->  :;<=>?@
Data &B01000111 , &B01110011 , &B00101101 , &B00110001 , &B01101101 , &B01001100 , 0

'Ascii 65 -  90        ABCDEFGHIJKLMNOPQRSTUVWXYZ
Data &B00000110 , &B00010001 , &B00010101 , &B00001001 , &B00000010 , &B00010100 , &B00001011 , &B00010000 , &B00000100
Data &B00011110 , &B00001101 , &B00010010 , &B00000111 , &B00000101 , &B00001111 , &B00010110 , &B00011011 , &B00001010
Data &B00001000 , &B00000011 , &B00001100 , &B00011000 , &B00001110 , &B00011001 , &B00011101 , &B00010011

'Ascii 91 - 127  ->  Muell